
;; $Revision: 1.4 $ $Date: 2005/06/14 21:05:45 $
(require-extension srfi-37)

;;; macro: (args:make-option (OPTION-NAME ...) ARG-DATA [BODY])

;; OPTION-NAME ... is a sequence of short or long option names.  They must be literal
;; symbols; single-character symbols become short options, and longer symbols become
;; long options.  So (args:make-option (c cookie) <...>) specifies a short option -c
;; and long option --cookie.  Underneath, (c cookie) becomes '(#\c "cookie"), as
;; expected by SRFI 37's OPTION.
;; 
;; ARG-DATA is either a pair (ARG-TYPE ARG-NAME) or a plain keyword ARG-TYPE.
;; ARG-TYPE is a keyword that specifies whether the option takes an argument:
;;    #:required   Argument is required
;;    #:optional   Argument is optional
;;    #:none       Does not take an argument (actually, any other value than
;;                 #:required or #:optional is interpreted as #:none)
;; ARG-NAME, if provided, is a string specifying the name of the argument.
;; This name is used in the help text produced by args:usage.
;;
;; BODY is an optional sequence of statements executed when this option is encountered.
;; Behind the scenes, BODY is wrapped in code which adds the current option and its
;; argument to the final options alist.  So, simply leave BODY blank and options
;; will be collected for you.  BODY is an option-processor as defined in SRFI 37,
;; and has access to the variables OPT (the current #<option>), NAME (the option name)
;; and ARG (argument value or #f).

(define-macro (args:make-option names arg-data docstring . body)
  (let* ((srfi37-names (map (lambda (name)
                              (let ((str (symbol->string name)))
                                (if (= (string-length str) 1)
                                    (string-ref str 0)
                                    str)))
                            names))
         (arg-name (if (pair? arg-data) (cadr arg-data) "ARG"))
         (arg-type (if (pair? arg-data) (car arg-data) arg-data)))
    `(make-args:option
      (option ',srfi37-names
              ,(eq? arg-type #:required)
              ,(eq? arg-type #:optional)
              (lambda (opt name arg options operands)
                ,@body
                (values (append (map (lambda (n) (cons n arg))
                                     ',names)
                                options)
                        operands)))
                ;;(values (cons (cons name arg) options) operands)))
      ,arg-name
      ,docstring)))
  
;;; macro: (args:make-operand-proc [BODY])

;; Return a procedure suitable for using as an operand procedure
;; in args:parse.  Provides the arguments OPERAND, OPTIONS, and OPERANDS
;; to the BODY; where OPERAND is the current operand (as in args-fold)
;; and OPTIONS and OPERANDS are SEEDS (as in args-fold) and should not be modified.
;; Also wraps BODY in code that adds the operand to the final operand list (seed).
(define-macro (args:make-operand-proc . body)
  `(lambda (operand options operands)
     ,@body
     (values options (cons operand operands))))
